import pytest
from django.core.exceptions import ValidationError
from django.db import IntegrityError
from obapi.models import Idea, Topic
from obapi.models.classifiers import IdeaAlias, TopicAlias


@pytest.mark.django_db
class TestCreateAliasedModel:
    def test_create_valid_object_with_two_aliases(self):
        # Act
        Topic.objects.create(name="Law&Øther", description="Law and other topics.")

        # Assert
        obj = Topic.objects.get(name="Law&Øther")
        assert obj.name == "Law&Øther"
        assert obj.slug == "law-other"
        assert obj.description == "Law and other topics."

        obj_aliases = obj.aliases.all()
        expected_aliases = {"law-other", "Law&Øther"}
        assert set(obj_aliases.values_list("text", flat=True)) == expected_aliases
        assert all(obj_aliases.values_list("protected", flat=True))

    def test_create_valid_model_with_one_alias(self):
        # Act
        Topic.objects.create(name="topic-two")  # slug is same as name

        # Assert
        obj = Topic.objects.get(name="topic-two")
        assert obj.name == "topic-two"
        assert obj.slug == obj.name
        assert obj.description == ""

        obj_aliases = obj.aliases.all()
        expected_aliases = {"topic-two"}
        assert set(obj_aliases.values_list("text", flat=True)) == expected_aliases
        assert obj_aliases[0].protected

    @pytest.mark.parametrize(
        "new_name,new_slug",
        [
            ("Law&Other", "law-other"),
            ("law other", "law-other"),
            ("law-other", "law-other"),
            ("LawOther", "lawother"),
        ],
    )
    def test_change_object_name(self, new_name, new_slug):
        # Act
        example = Topic.objects.create(
            name="Law&Øther", description="Law and other topics."
        )
        example.name = new_name
        example.save()

        # Assert
        assert example.name == new_name
        assert example.slug == new_slug

    def test_can_create_another_object_type_with_same_name(self):
        # Act
        topic = Topic.objects.create(name="Duplicated", description="A duplicate.")
        idea = Idea.objects.create(name="Duplicated", description="A duplicate.")

        # Assert
        assert topic.name == idea.name
        assert topic.slug == idea.slug
        assert set(topic.aliases.values_list("text", flat=True)) == set(
            idea.aliases.values_list("text", flat=True)
        )

    def test_cannot_duplicate_names_or_aliases(self):
        # Arrange
        topic = Topic.objects.create(name="Law&Other", description="About law.")
        topic.aliases.create(text="law-etc")

        # Act & assert
        # Same name and description
        with pytest.raises(IntegrityError):
            Topic.objects.create(name="Law&Other", description="About law.")

        # Different name, same slug
        with pytest.raises(IntegrityError):
            Topic.objects.create(name="Law Other")

        # Name = existing alias
        with pytest.raises(IntegrityError):
            Topic.objects.create(name="law-etc")

        # Slug = existing alias
        with pytest.raises(IntegrityError):
            Topic.objects.create(name="Law, Etc.")


@pytest.mark.django_db
class TestValidateUniqueAlias:
    def test_succeeds_for_unique_aliases(self):
        # Arrange
        law = Topic.objects.create(name="Law", description="About law.")
        law.aliases.create(text="Legal")
        norms = Topic.objects.create(name="Norms", description="About norms.")
        social = TopicAlias(text="Social Norms", owner=norms)

        # Act & assert
        assert social.validate_unique() is None  # not yet saved
        social.save()
        assert social.validate_unique() is None

    def test_succeeds_when_another_object_type_has_same_alias(self):
        # Arrange
        law_topic = Topic.objects.create(name="Law", description="About law.")
        law_topic.aliases.create(text="Rules")
        norms_idea = Idea.objects.create(name="Norms", description="About norms.")
        rules = IdeaAlias(text="Rules", owner=norms_idea)

        # Act & assert
        assert rules.validate_unique() is None  # not yet saved
        rules.save()
        assert rules.validate_unique() is None

    @pytest.mark.parametrize("alias_text", ["Law", "law", "Rules", "rulEs"])
    def test_fails_with_duplicate_aliases(self, alias_text):
        # Arrange
        law = Topic.objects.create(name="Law", description="About law.")
        law.aliases.create(text="Rules")
        norms = Topic.objects.create(name="Norms", description="About norms.")

        # Assert
        norms_alias = TopicAlias(text=alias_text, owner=norms)
        with pytest.raises(ValidationError):
            norms_alias.validate_unique()



from django.contrib import admin, messages
from django.contrib.admin.helpers import Fieldset
from django.contrib.admin.templatetags.admin_urls import add_preserved_filters
from django.contrib.admin.utils import model_ngettext
from django.core.exceptions import (
    FieldDoesNotExist,
    PermissionDenied,
    SuspiciousOperation,
)
from django.db import IntegrityError, transaction
from django.db.models import Q
from django.forms import BaseInlineFormSet, ValidationError
from django.http import Http404, HttpResponseRedirect
from django.template.response import TemplateResponse
from django.urls import path, reverse
from django.utils.translation import ngettext

from obapi import utils
from obapi.exceptions import APICallError
from obapi.forms import (
    AddOBContentItemForm,
    AddSpotifyContentItemForm,
    AddYoutubeContentItemForm,
)
from obapi.models import (
    Author,
    AuthorAlias,
    ContentItem,
    Idea,
    IdeaAlias,
    OBContentItem,
    SpotifyContentItem,
    Tag,
    TagAlias,
    Topic,
    TopicAlias,
    YoutubeContentItem,
)
from obapi.models.classifiers import ExternalLink

# Inlines
# https://docs.djangoproject.com/en/4.0/ref/contrib/admin/#inlinemodeladmin-objects


class AliasInlineFormSet(BaseInlineFormSet):
    """Inline which prevents duplicate aliases."""

    def clean(self):
        super().clean()

        if self.forms == [] or not hasattr(self, "cleaned_data"):
            return

        name = self.data["name"]
        (name_lower, slug) = (name.lower(), utils.to_slug(name))
        aliases_lower = [
            form["text"].data.lower() for form in self.forms if not form["DELETE"].data
        ]

        # Check aliases aren't the same as each other, or the author name / slug
        for i, alias_lower in enumerate(aliases_lower):

            # Alias is same as existing name / slug - not ok
            if alias_lower in (name_lower, slug):
                # Unless name is changed to existing alias
                if not self.forms[i].has_changed():
                    continue

                if self.instance.pk is not None:
                    self.instance.refresh_from_db(fields=["name"])
                raise ValidationError("Aliases must differ from name.", code="invalid")

            # Any duplication of aliases - not ok
            if any(alias_lower == aliases_lower[k] for k in range(0, i)):
                if self.instance.pk is not None:
                    self.instance.refresh_from_db(fields=["name"])
                raise ValidationError("Aliases must be unique.", code="invalid")


class AliasInline(admin.TabularInline):
    extra = 0
    formset = AliasInlineFormSet

    def get_queryset(self, request):
        return super().get_queryset(request).filter(protected=False)


class AuthorAliasInline(AliasInline):
    model = AuthorAlias


class IdeaAliasInline(AliasInline):
    model = IdeaAlias


class TopicAliasInline(AliasInline):
    model = TopicAlias


class TagAliasInline(AliasInline):
    model = TagAlias


class ContentItemInline(admin.StackedInline):
    model = ContentItem.external_links.through
    extra = 0

    autocomplete_fields = ("contentitem",)


# Admin actions
# https://docs.djangoproject.com/en/4.0/ref/contrib/admin/actions/


@admin.action(description="Merge objects", permissions=["change"])
def merge_aliased_objects(modeladmin, request, queryset):
    """Merge multiple objects and their aliases."""
    if queryset.count() < 2:
        modeladmin.message_user(
            request,
            "Two or more items must be selected to merge. No items have been changed.",
            messages.WARNING,
        )
        return

    aliased_model = queryset.model
    new_fields = {}

    # Make list of aliases
    aliases = []
    for obj in queryset:
        aliases.extend(obj.aliases.values_list("text", flat=True))
    new_fields["name"] = aliases.pop(0)

    # Create new description (if description field exists)
    try:
        max_length = aliased_model._meta.get_field("description").max_length
        description_list = queryset.values_list("description", flat=True)
        description = "/".join([desc for desc in description_list if desc != ""])
        new_fields["description"] = description[0:max_length]
    except FieldDoesNotExist:
        pass

    # Delete old objects and create new one (and its aliases)
    with transaction.atomic():
        queryset.delete()
        new_object = aliased_model.objects.create(**new_fields)
        for alias in aliases:
            new_object.aliases.create(text=alias)

    modeladmin.message_user(
        request,
        f"{len(aliases) + 1} aliases merged into {new_object.name}.",
        messages.SUCCESS,
    )


def convert_aliased_object(obj, model):
    """Convert an aliased object to another type."""
    with transaction.atomic():
        # Ensure uniqueness of aliases
        aliases = list(obj.aliases.values_list("text", flat=True))
        q_list = Q()
        for q in [Q(alias__text__iexact=alias) for alias in aliases]:
            q_list |= q
        if model.objects.filter(q_list).exists():
            raise IntegrityError

        # Create new object
        new_object = model.objects.create(name=obj.name)
        aliases.remove(new_object.name)
        for alias in aliases:
            new_object.aliases.create(text=alias)

        # Update related content
        items = obj.content.all()
        new_object.content.add(*items)

        # Delete old object
        obj.delete()
        del obj


def convert_aliased_objects(modeladmin, request, queryset, model):
    # Track errors and successes
    err_count = success_count = 0

    # Attempt to convert each object in the queryset
    for obj in queryset:
        try:
            convert_aliased_object(obj, model)
            success_count = success_count + 1
        except IntegrityError:
            err_count = err_count + 1
            pass

    # Message user about errors and successes
    if success_count > 0:
        modeladmin.message_user(
            request,
            f"{success_count} {model._meta.verbose_name_plural} successfully created",
            messages.SUCCESS,
        )
    if err_count > 0:
        modeladmin.message_user(
            request,
            f"{err_count} {model._meta.verbose_name_plural} could not be created",
            messages.WARNING,
        )


@admin.action(description="Convert selected items to Authors")
def convert_to_authors(modeladmin, request, queryset):
    convert_aliased_objects(modeladmin, request, queryset, Author)


@admin.action(description="Convert selected items to Ideas")
def convert_to_ideas(modeladmin, request, queryset):
    convert_aliased_objects(modeladmin, request, queryset, Idea)


@admin.action(description="Convert selected items to Topics")
def convert_to_topics(modeladmin, request, queryset):
    convert_aliased_objects(modeladmin, request, queryset, Topic)


@admin.action(description="Convert selected items to Tags")
def convert_to_tags(modeladmin, request, queryset):
    convert_aliased_objects(modeladmin, request, queryset, Tag)


# Model Admins
# https://docs.djangoproject.com/en/4.0/ref/contrib/admin/#modeladmin-objects


@admin.register(Author)
class AuthorAdmin(admin.ModelAdmin):
    inlines = [AuthorAliasInline]
    actions = [
        merge_aliased_objects,
        convert_to_ideas,
        convert_to_topics,
        convert_to_tags,
    ]
    search_fields = ("name",)


@admin.register(Idea)
class IdeaAdmin(admin.ModelAdmin):
    inlines = [IdeaAliasInline]
    actions = [
        merge_aliased_objects,
        convert_to_authors,
        convert_to_topics,
        convert_to_tags,
    ]
    search_fields = ("name",)


@admin.register(Topic)
class TopicAdmin(admin.ModelAdmin):
    inlines = [TopicAliasInline]
    actions = [
        merge_aliased_objects,
        convert_to_authors,
        convert_to_ideas,
        convert_to_tags,
    ]
    search_fields = ("name",)


@admin.register(Tag)
class TagAdmin(admin.ModelAdmin):
    inlines = [TagAliasInline]
    actions = [
        merge_aliased_objects,
        convert_to_authors,
        convert_to_ideas,
        convert_to_topics,
    ]
    search_fields = ("name",)


@admin.register(ExternalLink)
class ExternalLinkAdmin(admin.ModelAdmin):
    inlines = [
        ContentItemInline,
    ]
    exclude = ("external_links",)
    search_fields = ("url",)
    actions = ("delete_unreferenced_links",)

    @admin.action(description="Delete unreferenced links", permissions=["delete"])
    def delete_unreferenced_links(self, request, queryset):
        delete_count, details = queryset.filter(content__isnull=True).delete()
        self.message_user(
            request,
            f"{delete_count} links deleted",
            messages.INFO,
        )


class ContentItemAdminTemplate(admin.ModelAdmin):
    add_form_template = "admin/obapi/add_form.html"
    AddForm = None  # override in subclasses

    actions = ["update_selected_items", "internalize_links"]

    readonly_fields = ("create_timestamp", "update_timestamp", "download_timestamp")

    search_fields = ("title",)

    autocomplete_fields = (
        "authors",
        "ideas",
        "topics",
        "tags",
        "external_links",
        "internal_links",
    )

    def add_view(self, request, form_url="", extra_context=None):
        if not self.has_add_permission(request):
            raise PermissionDenied
        if request.method == "POST":
            form = self.AddForm(request.POST)
            if form.is_valid():
                id_field = next(iter(form.fields.values())).label
                message_if_error = (
                    f"Could not retrieve item. Please check the {id_field} is valid."
                )
                try:
                    item = form.save()
                except APICallError:
                    item = None
                except IntegrityError:
                    item = None
                    message_if_error = "The specified item already exists."
                if item is not None:
                    change_message = f"Created item {item}."
                    self.log_addition(request, item, change_message)
                    return self.response_add(request, item)

                self.message_user(
                    request,
                    message_if_error,
                    messages.ERROR,
                )

        else:
            form = self.AddForm()

        opts = self.model._meta
        fieldset = Fieldset(form, fields=form.fields, model_admin=self)
        context = {
            **self.admin_site.each_context(request),
            "title": f"Add {self.model._meta.verbose_name}",
            "fieldset": fieldset,
            "media": self.media,
            "errors": form.errors,
            "add": True,
            "has_view_permission": self.has_view_permission(request, None),
            "has_add_permission": self.has_add_permission(request),
            "has_change_permission": self.has_change_permission(request, None),
            "has_delete_permission": self.has_delete_permission(request, None),
            "opts": opts,
            "app_label": opts.app_label,
            **(extra_context or {}),
        }
        request.current_app = self.admin_site.name
        form_template = self.add_form_template
        return TemplateResponse(request, form_template, context)

    @admin.action(description="Update selected items", permissions=["change"])
    def update_selected_items(self, request, queryset):
        queryset.update_items()

    @admin.action(description="Internalize links", permissions=["change"])
    def internalize_links(self, request, queryset):
        try:
            for item in queryset:
                item.internalize_links(clear=False)
        except IntegrityError:
            self.message_user(
                request, "Error when internalizing links.", messages.ERROR
            )
        else:
            self.message_user(
                request, "Links internalized successfully.", messages.SUCCESS
            )


@admin.register(ContentItem)
class ContentItemAdmin(ContentItemAdminTemplate):
    # Revert to standard add view
    def add_view(self, request, form_url="", extra_context=None):
        return self.changeform_view(request, None, form_url, extra_context)

    @admin.action(description="Update selected items", permissions=["change"])
    def update_selected_items(self, request, queryset):
        for model in (YoutubeContentItem, SpotifyContentItem, OBContentItem):
            model.objects.filter(
                pk__in=queryset.select_subclasses(model)
            ).update_items()


@admin.register(YoutubeContentItem)
class YoutubeContentItemAdmin(ContentItemAdminTemplate):
    AddForm = AddYoutubeContentItemForm


@admin.register(SpotifyContentItem)
class SpotifyContentItemAdmin(ContentItemAdminTemplate):
    AddForm = AddSpotifyContentItemForm


@admin.register(OBContentItem)
class OBContentItemAdmin(ContentItemAdminTemplate):
    AddForm = AddOBContentItemForm

    def get_urls(self):
        urls = super().get_urls()
        custom_urls = [
            path(
                "update/",
                self.admin_site.admin_view(self.update_view),
                name="obapi_obcontentitem_update",
            )
        ]
        return custom_urls + urls

    def changelist_view(self, request, extra_context=None):
        if extra_context is None:
            extra_context = {}
        extra_context.update(has_change_permission=self.has_change_permission(request))
        return super().changelist_view(request, extra_context)

    def update_view(self, request):
        """View to update existing OB content items.

        Either downloads all new posts ("pull") or updates all edited posts ("sync").

        Works by (1) performing the requested operation, (2) logging the results and
        messaging the user, and (3) redirecting to list / index page.
        """
        # Accept POST requests only
        if request.method != "POST":
            raise Http404

        if "_pull" in request.POST:
            # Check permissions
            if not self.has_add_permission(request):
                raise PermissionDenied
            # Pull items
            created_items = OBContentItem.objects.download_new_items()
            # Messaging & logging
            changecount = len(created_items)
            changetype = "created"
            for item in created_items:
                self.log_addition(request, item, f"Created item {item}.")
        elif "_sync" in request.POST:
            # Check permissions
            if not self.has_change_permission(request):
                raise PermissionDenied
            # Sync items
            updated_items = OBContentItem.objects.update_edited_items()
            # Messaging & logging
            changecount = len(updated_items)
            changetype = "updated"
            for item in updated_items:
                self.log_change(request, item, f"Updated item {item}.")
        else:
            raise SuspiciousOperation

        # Send message
        if changecount == 0:
            msg = f"No items were {changetype}."
            status = messages.INFO
        else:
            msg = ngettext(
                "Successfully %(changetype)s %(changecount)s %(name)s.",
                "Successfully %(changetype)s %(changecount)s %(name)s.",
                changecount,
            ) % {
                "changetype": changetype,
                "changecount": changecount,
                "name": model_ngettext(self.model._meta, changecount),
            }
            status = messages.SUCCESS
        self.message_user(request, msg, status)

        # Redirect logic - copied from internal method `_response_post_save`
        opts = self.model._meta
        if self.has_view_or_change_permission(request):
            post_url = reverse(
                "admin:%s_%s_changelist" % (opts.app_label, opts.model_name),
                current_app=self.admin_site.name,
            )
            preserved_filters = self.get_preserved_filters(request)
            post_url = add_preserved_filters(
                {"preserved_filters": preserved_filters, "opts": opts}, post_url
            )
        else:
            post_url = reverse("admin:index", current_app=self.admin_site.name)
        return HttpResponseRedirect(post_url)



import pytest
from obapi.utils import parse_duration


class TestParseDuration:
    def test_gives_correct_result_for_valid_durations(self):
        assert parse_duration("PT10S").total_seconds() == pytest.approx(10)
        assert parse_duration("P3DT3H5M2S").total_seconds() == pytest.approx(270302)

    def test_raises_error_for_invalid_durations(self):
        with pytest.raises(ValueError):
            parse_duration("arbitrary string")

        with pytest.raises(ValueError):
            parse_duration("PT3.827M")



from django.core.exceptions import ValidationError
from django.db import IntegrityError, models, transaction
from django.db.models import UniqueConstraint
from django.db.models.functions import Lower
from django.urls import reverse
from obapi import utils


class AliasedModel(models.Model):
    """Base class for models with aliases."""

    name = models.CharField(max_length=100, unique=True, help_text="Name.")
    slug = models.SlugField(max_length=100, unique=True, null=True, editable=False)
    description = models.CharField(
        max_length=100, help_text="Brief description.", blank=True
    )

    class Meta:
        abstract = True

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        """Save an AliasedModel instance.

        Raises
        ------
        IntegrityError
            If the instance `name` is already an alias of another instance.
        """
        try:
            with transaction.atomic():
                # (1) Update slug
                self.slug = utils.to_slug(self.name)
                # (2) Save model instance
                super().save(*args, **kwargs)
                # (3) un-protect all aliases
                self.aliases.all().update(protected=False)
                # (4) create and protect <instance name> alias
                self.aliases.update_or_create(
                    text__iexact=self.name,
                    defaults={"text": self.name, "protected": True},
                )
                # (5) create and protect <slug> alias (if needed)
                self.aliases.update_or_create(
                    text__iexact=self.slug,
                    defaults={"text": self.slug, "protected": True},
                )
        except IntegrityError as e:
            raise IntegrityError(
                f"The name {self.name}"
                f" is an alias of another {self._meta.verbose_name}."
            ) from e

    def get_absolute_url(self):
        model_name = self._meta.verbose_name
        return reverse(
            "explore_detail",
            kwargs={"model_name": model_name, "instance_name": self.slug},
        )


class Author(AliasedModel):
    description = None


class Idea(AliasedModel):
    pass


class Topic(AliasedModel):
    pass


class Tag(AliasedModel):
    description = None


class Alias(models.Model):
    """Base class for models which represent aliases of other models.

    `owner` is the Foreign Key to the aliased model. The aliased model should have a
    `name` attribute, or the subclass should override the __str__ method.
    """

    text = models.CharField(max_length=100, help_text="Alias text.")
    protected = models.BooleanField(
        default=False,
        editable=False,
        help_text="True for the alias which equals the owner name.",
    )
    owner = None

    class Meta:
        abstract = True
        verbose_name_plural = "aliases"
        constraints = [UniqueConstraint(Lower("text"), name="%(class)s_is_unique")]

    def __str__(self):
        return f"{self.text} ({self.owner.name})"

    def validate_unique(self, exclude=None):
        super().validate_unique(exclude=exclude)
        if exclude is not None and "text" in exclude:
            return
        # Raise error if there is a matching alias, and the match has a different pk
        try:
            match = type(self).objects.get(text__iexact=self.text)
        except type(self).DoesNotExist:
            return
        else:
            if match != self:
                self.owner.refresh_from_db(fields=["name"])
                raise ValidationError(
                    {"text": "This alias is already taken."}, code="invalid"
                )


class AuthorAlias(Alias):
    owner = models.ForeignKey(
        Author,
        verbose_name="author",
        on_delete=models.CASCADE,
        related_name="aliases",
        related_query_name="alias",
        help_text="Author the alias refers to.",
    )


class IdeaAlias(Alias):
    owner = models.ForeignKey(
        Idea,
        verbose_name="idea",
        on_delete=models.CASCADE,
        related_name="aliases",
        related_query_name="alias",
        help_text="Idea the alias refers to.",
    )


class TopicAlias(Alias):
    owner = models.ForeignKey(
        Topic,
        verbose_name="topic",
        on_delete=models.CASCADE,
        related_name="aliases",
        related_query_name="alias",
        help_text="Topic the alias refers to.",
    )


class TagAlias(Alias):
    owner = models.ForeignKey(
        Tag,
        verbose_name="tag",
        on_delete=models.CASCADE,
        related_name="aliases",
        related_query_name="alias",
        help_text="Tag the alias refers to.",
    )


class ExternalLink(models.Model):
    """Model representing link to "external" content.

    Here "external" means "outside the database".
    """

    url = models.URLField(unique=True, help_text="Link URL.")

    def __str__(self):
        return self.url



from django.conf import settings
from django.db import models
from ordered_model.models import OrderedModel


class Sequence(models.Model):
    owner = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    title = models.CharField(max_length=100, help_text="Sequence title.")
    description = models.TextField(
        blank=True, max_length=5000, help_text="Description of sequence."
    )
    items = models.ManyToManyField("ContentItem", through="SequenceMember")
    public = models.BooleanField(
        default=False, help_text="Whether the sequence is public or private."
    )

    def __str__(self):
        return self.title


class SequenceMember(OrderedModel):
    content_item = models.ForeignKey(
        "ContentItem",
        on_delete=models.CASCADE,
        related_name="sequence_members",
        related_query_name="sequence_members",
    )
    sequence = models.ForeignKey(
        Sequence,
        on_delete=models.CASCADE,
        related_name="members",
        related_query_name="members",
    )
    order_with_respect_to = "sequence"

    class Meta:
        ordering = ("sequence", "order")

    def __str__(self):
        return f"{self.content_item.title} ({self.sequence.title})"



import datetime

import pytest
from obapi.download import download_spotify_episodes_json, download_youtube_videos_json
from obapi.tidy import (
    _tidy_ob_post_object,
    _tidy_spotify_episode_json,
    _tidy_youtube_video_json,
)
from obscraper import get_post_by_url

from markers import require_spotify_api_auth, require_youtube_api_key


class TestTidyYoutubeVideoJSON:
    @require_youtube_api_key
    def test_returns_expected_result_for_example_videos(self):
        # Arrange
        tricks_response = download_youtube_videos_json(["C-gEQdGVXbk"])

        # Act
        tricks_json = _tidy_youtube_video_json(tricks_response["items"][0])

        # Assert
        assert (
            tricks_json["title"] == "10 Python Tips and Tricks For Writing Better Code"
        )
        assert tricks_json["duration"].total_seconds() == pytest.approx(2361)
        assert tricks_json["yt_description"].startswith("This video is")

        # Arrange
        trends_response = download_youtube_videos_json(["e6LOWKVq5sQ"])

        # Act
        trends_json = _tidy_youtube_video_json(trends_response["items"][0])

        # Assert
        assert trends_json["yt_description"] == ""
        assert set(trends_json["classifier_names"]) == {"simpsons", "disco stu"}
        assert trends_json["author_names"] == ["dumbmatter"]


class TestTidySpotifyEpisodeJSON:
    @require_spotify_api_auth
    def test_returns_expected_result_for_example_episode(self):
        # Arrange
        signals_response = download_spotify_episodes_json(["6MAszRR6tdDnMsjgVdw4Jh"])

        # Act
        signals_json = _tidy_spotify_episode_json(signals_response["episodes"][0])

        # Assert
        assert (
            signals_json["title"]
            == "Robin Hanson on Signaling and Self-Deception (Live at Mason Econ)"
        )
        assert signals_json["sp_show_title"] == "Conversations with Tyler"
        assert signals_json["sp_description"].startswith(
            "If intros aren’t about introductions"
        )


class TestTidyOBPostObject:
    def test_returns_expected_result_for_example_post(self):
        # Arrange
        post_raw = get_post_by_url(
            "https://www.overcomingbias.com/2009/03/signaling-in-economics.html"
        )

        # Act
        post_tidy = _tidy_ob_post_object(post_raw)

        # Assert
        assert post_tidy["title"] == "Signaling in Economics"
        assert post_tidy["publish_date"] == datetime.datetime(
            2009, 3, 21, 22, 0, tzinfo=datetime.timezone.utc
        )
        assert post_tidy["ob_post_number"] == 16642
        assert post_tidy["text_plain"].startswith("Arnold Kling cites this")



import datetime
import re

import bleach
from slugify import slugify

ISO_8601_DURATION_PATTERN = re.compile(
    (
        r"^P"  # starts with a P
        r"((?P<days>\d+)D)?"  # optional day field must come before T
        r"(T(?=\d+[HMS])((?P<hours>\d+)H)?((?P<minutes>\d+)M)?((?P<seconds>\d+)S)?)?"
        r"$"
    )
)


def parse_duration(duration):
    if (components := ISO_8601_DURATION_PATTERN.search(duration)) is None:
        raise ValueError(f'"{duration}" is not a valid ISO-8601 duration.')
    components = {
        k: int(v if v is not None else 0) for k, v in components.groupdict().items()
    }
    return datetime.timedelta(**components)


def plaintext_to_html(text):
    # Convert line breaks to <br> elements
    broken_text = text.replace("\n", "<br>")

    # Linkify URLs and email addresses
    return bleach.linkify(broken_text, parse_email=True)


def to_slug(text):
    return slugify(text)
